-module(bots_client).
-author('luke@bluetail.com').

% -define(debug,true).
% -define(logging,true).

-ifdef(debug).
-define(trace(Fmt, Args), io:format(Fmt, Args)).
-else.
-define(trace(Fmt, Args), true).
-endif.

-ifdef(logging).
-define(log(E), E).
-else.
-define(log(E), true).
-endif.

-import(lists, [foreach/2, splitwith/2, map/2, reverse/1, flatten/1]).

-export([run/1, test/0, live/1, start_link/2]).
-export([init/2]).			% internal

test() ->
    start_link("localhost", 10000).

live(N) ->
    start_link("icfp1.cse.ogi.edu", 22100 + N).

run([HostA, PortA]) ->
    start_link(atom_to_list(HostA),
	       list_to_integer(atom_to_list(PortA))).

start_link(Host, Port) ->
    proc_lib:spawn_link(?MODULE, init, [Host, Port]).

init(Host, Port) ->
    ?log(file:delete("trace.txt")),
    {ok, S} = gen_tcp:connect(Host, Port, [{packet, line},
					   {active, false},
					   list]),
    send_line(S, "Player"),
    read_board(S),
    {ID, Capacity, Money} = read_player_configuration(S),
    bots_db:set_configuration(ID, Capacity, Money),
    handle_server_reponse(S),
    loop(S, bots_player:init(ID, Capacity, Money)).

%% ------------------------------------------------------------
%% Initial board reading

read_board(S) ->
    {W, H} = read_dimensions(S),
    bots_db:new_map(W, H),
    read_rows(S, H).

read_dimensions(S) ->
    [W, H] = split(read_line(S)),
    {list_to_integer(W), list_to_integer(H)}.

read_rows(S, 0) ->
    bots_db:end_map();
read_rows(S, N) when N > 0 ->
    bots_db:add_map_line(read_line(S)),
    read_rows(S, N-1).

%% ------------------------------------------------------------
%% Initial player configuration reading

read_player_configuration(S) ->
    [ID, Capacity, Money] = [list_to_integer(X)
			     || X <- split(read_line(S))],
    {ID, Capacity, Money}.

%% handle_server_reponse(S) -> ok
handle_server_reponse(S) ->
    Events = read_server_response(S),
    LiveBots = [Bot || {Bot, _} <- Events],
    bots_db:set_robot_list(LiveBots),
    foreach(fun robot_event_update_db/1, Events),
    ok.

%% read_server_response(S) -> [{RobotID, [Action]}]
%% Events = [{RobotID, [Action]}]
%% Packages = [PackageID]
%% Action = {move, Direction}
%%        | {pick, PackageID}
%%        | {drop, PackageID}
%%        | {position, {X, Y}}
%% RobotID = PackageID = integer()
%% Direction = north | south | east | west
read_server_response(S) ->
    parse_all_actions(split(read_line(S))).

%% Read list of dropped packages at our location, updated the DB, and
%% return them.
read_and_update_packages(S) ->
    Packages = read_packages(S),
    foreach(fun({ID, X, Y, Weight}) ->
		    bots_db:you_have_stood_on_a_package(ID, X, Y, Weight)
	    end,
	    Packages),
    Packages.

parse_all_actions([]) ->
    [];
parse_all_actions(["#"++RobotID|Tokens]) ->
    {This, Rest} = splitwith(fun("#"++_) -> false;
				(_)      -> true
			     end,
			     Tokens),
    [{list_to_integer(RobotID), parse_actions(This)} |
     parse_all_actions(Rest)].

parse_actions([]) ->
    [];
parse_actions(["N" | Rest]) -> [{move, north} | parse_actions(Rest)];
parse_actions(["S" | Rest]) -> [{move, south} | parse_actions(Rest)];
parse_actions(["E" | Rest]) -> [{move, east}  | parse_actions(Rest)];
parse_actions(["W" | Rest]) -> [{move, west}  | parse_actions(Rest)];
parse_actions(["P", ID | Rest]) ->
    [{pick, list_to_integer(ID)} | parse_actions(Rest)];
parse_actions(["D", ID | Rest]) ->
    [{drop, ID} | parse_actions(Rest)];
parse_actions(["X", X, "Y", Y | Rest]) ->
    [{position, {list_to_integer(X), list_to_integer(Y)}} |
     parse_actions(Rest)].

read_packages(S) ->
    parse_packages(split(read_line(S))).

parse_packages([]) ->
    [];
parse_packages([ID, X, Y, Weight | Rest]) ->
    [list_to_tuple(map(fun(L) -> list_to_integer(L) end,
		       [ID, X, Y, Weight]))
     | parse_packages(Rest)].

%% ------------------------------------------------------------
%% Main loop

loop(Sock, State) ->
    Packages = read_and_update_packages(Sock),
    ?log(write_map_trace()),
    State1 = case catch bots_player:turn(State, Packages) of
		 {Bid, Move, NewState} ->
		     send_move(Sock, Bid, Move),
		     NewState;
		 {'EXIT', Rsn} ->
		     ?trace("bots_player crashed: ~n  ~p~n", [Rsn]),
		     send_move(Sock, 1, nop),
		     bots_player:reinit(State)
	     end,
    handle_server_reponse(Sock),
    loop(Sock, State1).

robot_event_update_db({ID, Actions}) ->
    foreach(fun({position, {X,Y}}) ->
		    bots_db:set_robot_position(ID, X, Y);
	       ({move, Direction}) ->
		    bots_db:robot_moved(ID, Direction);
	       ({pick, PID}) ->
		    bots_db:robot_picked_up_package(ID, PID);
	       ({drop, PID}) ->
		    bots_db:robot_dropped_package(ID, PID)
	    end,
	    Actions).

write_map_trace() ->
    {ok, Fd} = file:open("trace.txt", [append]),
    io:format(Fd, "~n", []),
    bots_db:print_map(Fd),
    file:close(Fd).

%% ------------------------------------------------------------
%% Making a move

send_move(S, Bid, Move) ->
    send_line(S, [io_lib:format("~p ", [Bid]),
		  format_move(Move)]).

format_move({move, Direction}) ->
    io_lib:format("Move ~s",
		  [case Direction of
		       north -> "N";
		       south -> "S";
		       east  -> "E";
		       west  -> "W"
		   end]);
format_move({pick, Packages}) ->
    ["Pick", map(fun(P) -> io_lib:format(" ~p", [P]) end,
		 Packages)];
format_move({drop, Packages}) ->
    ["Drop", map(fun(P) -> io_lib:format(" ~p", [P]) end,
		 Packages)];
format_move(nop) ->
    "Drop -1".

%% ------------------------------------------------------------
%% Utility

send_line(S, Line) ->
    ?trace(">> ~s~n", [Line]),
    gen_tcp:send(S, [Line,"\n"]).

read_line(S) ->
    read_line(S, []).

read_line(S, Acc) ->
    {ok, Line} = gen_tcp:recv(S, 0),
    case lists:last(Line) of
	$\n ->
	    Final = flatten(reverse([(Line -- "\n")|Acc])),
	    ?trace("<< ~s~n", [Final]),
	    Final;
	_ ->				% line continues
	    read_line(S, [Line|Acc])
    end.

split(Line) ->
    string:tokens(Line, " ").
