%%% File    : bots_db.erl
%%% Author  : Joe Armstrong (joe@sics.se)
%%% Purpose : *imperative* data base of robot map etc. 

-module(bots_db).

-export([new_map/2, 
	 add_map_line/1,
	 end_map/0,
	 print_map/0,
	 print_map/1,
	 set_configuration/3,
	 set_dest/1,

	 all_robots/0,
	 all_pkgs/0,
	 all_homes/0,
	 all_dropped/0,
	 all_other_robot_positions/1,

	 info_robot/1,
	 info_pkg/1,
	 info_map/0,
	 info_cell/1,
	 info_config/0, 

	 set_robot_list/1,
	 set_robot_position/3,
	 set_robot_x/2,
	 set_robot_y/2,
	 robot_moved/2,
	 robot_picked_up_package/2,
	 robot_dropped_package/2,
	 you_have_stood_on_a_package/4]).

-import(lists, [delete/2, foreach/2, map/2, member/2]).

-define(WALL,  $#).
-define(WATER, $~).
-define(EMPTY, $.).
-define(HOME,  $@).

%% Interface:
%%
%%   new_map(Width, Height) -> true   -- Initialse map
%%   add_map_line(Str) -> true.       -- Add a line to the map
%%   end_map() -> true.               -- End of Map
%%
%%   all_robots()  -> [RobotId].
%%   all_homes()   -> [Pos]
%%   all_pkgs()    -> [PkgId]
%%   all_dropped() -> [PkgId]
%%
%%   info_robot(RobotId) -> {Pos, Money, [PkgId]}
%%   info_pkg()          -> {CurrentPos, Weight, DestPos, {robot,I} | dropped}
%%   info_map()          -> {Width, Height}
%%   info_cell({X, Y})   -> wall | water | adjacent 
%%   info_config()       -> {MyId, Cap, Money}
%%   set_robot_position(RobotId, X, Y) -> true
%%   set_robot_x(RobotId, X)           -> true
%%   set_robot_y(RobotId, Y)           -> true
%%   set_dest({Pos,Wt} | undefined, )  -> true.
%%
%%   robot_moved(RobotId, Direction)  -> true.
%%      Direction = north | south | east | west
%%
%%   robot_picked_up_package(RobotId, PkgId) -> true
%%   robot_dropped_package(RobotId, PkgId) -> true
%%
%%   a_package_has_been_found(PkgId, Weight, Dest) -> true
%%
%% Types
%%   Pos = {X, Y}, X=int() Y=int() 

%%  ets dta structures
%%  ------------------
%%   Key          Type
%%   ---          ----
%%   {X, Y}    => wall | water | adjacent
%%   {robot,I} => {Pos, [PkgId]}
%%   {pkg, I}  => {CurrentPos, Weight, DestPos, {robot,I} | dropped}
%%   robots    => [RobotId]
%%   dropped   => [PkgId]
%%   pkgs      => [PkgId]
%%   homes     => [Pos]
%%   config    => {MyId, MaxCarry, Money}
%%   dest      => Pos | undefined

%% To make a map
%%  call
%%    new_map(Width, Ht)
%%      add_map_line(Str1)
%%      add_map_line(Str2)
%%      ...
%%    end_map()

%%======================================================================

new_map(Width,  Height) ->	 
    new_table(),
    insert(size, {Width, Height}),
    insert(line, 1),
    insert(robots, []),
    insert(dropped, []),
    insert(homes, []),
    insert(pkgs, []),
    insert(dest, undefined),
    make_outer_wall(Width, Height).


%% The map coordinate system is like this:
%%
%% .............. (Width, Ht)
%% ..........................
%% (1,1) ....................


%% make_outer_wall 
%%  creates an outer wall that surrounds the entire map

make_outer_wall(Width, Height) ->
    for(0, Width+1,  fun(X)  -> set_type(X, 0, ?WALL) end),
    for(0, Width+1,  fun(X)  -> set_type(X, Height+1, ?WALL) end),
    for(0, Height+1, fun(Y)  -> set_type(0, Y, ?WALL) end),
    for(0, Height+1, fun(Y)  -> set_type(Width+1, Y, ?WALL) end).

add_map_line(Str) ->
    Ht = lookup(line),
    add_lines(1, Ht, Str),
    insert(line, Ht+1).

add_lines(X, Y, [H|T]) ->
    set_type(X, Y, H),
    add_lines(X+1, Y, T);
add_lines(X, Y, []) ->
    [].

set_type(X, Y, ?EMPTY) ->
    true;
set_type(X, Y, ?WATER) ->
    insert({X,Y}, water);
set_type(X, Y, ?WALL) ->
    insert({X,Y}, wall);
set_type(X, Y, $@) ->
    L = lookup(homes),
    insert(homes, [{X,Y}|L]).

end_map() ->
    %% Find all the walls
    Incs  = [{1,1},{1,0},{1,-1},{0,1},{0,-1},{-1,1},{-1,0},{-1,-1}],
    Water = all_water_pos(),
    foreach(fun(Pos) -> make_adjacent_to_water(Pos, Incs) end, Water).

make_adjacent_to_water({X, Y}, Incs) ->
    foreach(fun({I,J}) ->
		    Pos = {X+I, Y+J},
		    case find(Pos) of
			not_found ->
			    insert(Pos, adjacent);
			_ -> 
			    true
		    end
	    end, Incs).

set_configuration(Id, MaxCarry, Money) ->
    insert(config, {Id, MaxCarry, Money}).

set_dest(Dest) ->
    %%io:format("DDD set dest:~p~n",[Dest]),
    insert(dest, Dest).

all_homes()   -> lookup(homes).

all_dropped() -> lookup(dropped).

all_robots()  -> lookup(robots).

all_pkgs()    -> lookup(pkgs).


%% Set all living robot IDs (so that we prune dead ones)
set_robot_list(Ids) ->
    ets:insert(board, {robots, Ids}).

set_robot_position(Id, X, Y) ->
    set_robot_x(Id, X),
    set_robot_y(Id, Y).

set_robot_x(Id, X) ->
    ensure(robots, Id),
    case find(Key={robot, Id}) of
	{found, {{_,Y},Pkgs}}  ->
	    insert(Key, {{X,Y}, Pkgs});
	not_found ->
	    insert(Key, {{X, unknown}, []})
    end.

set_robot_y(Id, Y) ->
    ensure(robots, Id),
    case find(Key={robot, Id}) of
	{found, {{X,_},Pkgs}}  ->
	    insert(Key, {{X, Y}, Pkgs}); not_found ->
	    insert(Key, {{unknown, Y}, []})
    end.

robot_moved(Id, Dirn) ->
    case find(Key={robot, Id}) of
	{found, {{X, Y}, Pkgs}} ->
	    NewPos = case Dirn of 
			 north -> {X, Y+1};
			 south -> {X, Y-1};
			 east  -> {X+1, Y};
			 west  -> {X-1, Y}
		     end,
	    insert(Key, {NewPos, Pkgs});
	not_found  ->
	    exit({badarg, robot_moved, Id, Dirn})
    end.

robot_picked_up_package(RobotId, PkgId) ->
    %% Add to list of carried packages 
    {Pos, Pkgs} = lookup(Key={robot, RobotId}),
    insert(Key, {Pos,[PkgId|Pkgs]}),
    %% Have to remove from the dropped list
    D = lookup(dropped),
    insert(dropped, delete(PkgId, D)).

robot_dropped_package(RobotId, PkgId) ->
    %% remove from list of carried packages 
    {Pos, Pkgs} = lookup(Key={robot, RobotId}),
    insert(Key, {Pos, delete(PkgId, Pkgs)}),
    %% Remove from list of dropped packages
    D = lookup(dropped),
    insert(dropped, delete(PkgId, D)),
    %% remove package data
    ets_erase({pkg, PkgId}).

you_have_stood_on_a_package(Id, DestX, DestY, Weight) ->
    Pos = get_current_position(),
    ensure(pkgs, Id),
    ensure(dropped, Id),
    insert({pkg, Id}, {Pos, Weight, {DestX, DestY}, dropped}),
    %% remove Pos from the homes list (if any)
    Homes = all_homes(),
    Homes1 = delete(Pos, Homes),
    insert(homes, Homes1).

get_current_position() ->
    Id = get_my_id(),
    {Pos, _} = info_robot(Id),
    Pos.

%%----------------------------------------------------------------------

info_config() -> lookup(config).

info_map() -> lookup(size).

info_cell(Pos) ->
    case find(Pos) of
	{found, Obj} -> Obj;
	not_found -> empty
    end.

info_robot(Id) ->
    case find({robot, Id}) of
	{found,Facts} -> Facts;
	not_found -> exit({info_robot, Id})
    end.

info_pkg(Id) ->
    case find({pkg, Id}) of
	{found,Facts} -> Facts;
	not_found -> exit({info_pkg, Id})
    end.

%% Positions of all enemy robots.
all_other_robot_positions(MyId) ->
    lists:flatmap(fun(RID) ->
			  {Pos,_} = info_robot(RID),
			  if MyId == RID -> [];
			     true -> [Pos]
			  end
		  end,
		  all_robots()).

%%----------------------------------------------------------------------

get_my_id() ->
    {Id, _, _} = lookup(config),
    Id.

ensure(Key, Id) ->
    L = lookup(Key),
    case member(Id, L) of
	true  -> true;
	false -> insert(Key, [Id|L])
    end.

%%----------------------------------------------------------------------
%% print the map

print_map() ->
    print_map(standard_io).

print_map(Fd) ->
    {Width, Height} = info_map(),
    %% io:format(Fd, "Width=~p Ht=~p~n",[Width,Height]),
    downto(Height, 1,
	fun(I) ->
		Str = get_map_line(I, Width),
		io:format(Fd, "~s~n", [Str])
	end),
    {_,Capacity,_} = info_config(),
%    io:format(Fd, "Homes=~p~n",[all_homes()]),
%    io:format(Fd, "# Dropped=~p~n",[length(all_dropped())]),
    case lookup(dest) of
	undefined -> true;
	{_, Wt} ->
	    io:format(Fd, "Carrying:~p / ~p~n", [Wt, Capacity])
    end.


get_map_line(Height, Width) ->
    Homes = all_homes(),     %% [Pos]
    Dr     = all_dropped(), %% [PkgId]
    Robots = all_robots(),   %% [Id]
    MyId = get_my_id(),
    Dropped = map(fun(I) ->
			  {Pos,_,_,_} = info_pkg(I),
			  Pos
		  end, Dr),
    R1 = map(fun(I) -> 
		     {Pos,_} = info_robot(I),
		     {Pos, I}
	     end, Robots),
    map(1,Width,1,fun(I) ->
			  display(I,Height, Homes, Dropped, R1, MyId)
		  end).


display(X,Y,Homes,Dropped, Robs, MyId) ->
    Pos = {X, Y},
    DestPos  = case lookup(dest) of
		   undefined -> undefined;
		   {Pos1, _} -> Pos1
	       end,
    case Pos of
	DestPos ->
	    $D;
	_ ->
	    case member(Pos, Homes) of
		true -> 
		    ?HOME;
		false ->
		    case member(Pos, Dropped) of
			true ->
			    $d;
			false ->
			    case is_robot(Pos, Robs) of
				{yes, MyId} ->
				    $E;
				{yes, Id} ->
				    $Z;
				no ->
				    case info_cell(Pos) of
					wall  -> ?WALL;
					water -> ?WATER;
					_     -> ?EMPTY
						     end
			    end
		    end
	    end
    end.

is_robot(Pos, [{Pos,Id}|_]) -> {yes, Id};
is_robot(Pos, [_|T])        -> is_robot(Pos, T);
is_robot(Pos, [])           -> no.

%%----------------------------------------------------------------------
%% ets interface

lookup(Key) ->
    [{_,Val}] = ets:lookup(board, Key),
    Val.

find(Key) ->
    case ets:lookup(board, Key) of
	[{_,Val}] -> {found, Val};
	[] -> not_found
    end.

ets_erase(Key) ->
    ets:delete(board, Key).

insert(Key, Value) -> ets:insert(board, {Key, Value}).
	     
new_table()        -> ets:new(board, [set, protected, named_table]).

all_water_pos() ->
    L = ets:match(board, {'$1', water}),	
    map(fun([Pos]) -> Pos end, L).

%%----------------------------------------------------------------------
%% utilities

for(I, Max, F) when I > Max -> true;
for(I, Max, F)              -> F(I), for(I+1, Max, F).

downto(I, Min, F) when I < Min -> true;
downto(I, J,  F)               -> F(I), downto(I-1,J,F).

map(I,Max, J, F) when I > Max -> [];
map(I,Max,J, F)               -> [F(I)|map(I+1,Max, J,F)].

