%% Purpose: pattern matching of map fragments

-module(bots_match).

-export([match/4]).
-compile(export_all).
-import(lists, [member/2, reverse/1, foreach/2]).

%%-define(debug,true).

-ifdef(debug).
-define(trace(E), E).
-else.
-define(trace(E), true).
-endif.

%% Pattern recogniser context
-record(rec, {id,			% our id
	      enemies,			% [{X,Y}] bad guy positions
	      x0, y0,			% position of first pattern element
	      pos,			% our own position
	      next			% our next planned position
	     }).

%% Does the pattern match at our current location, and if so, where
%% does element {X,Y} of the pattern lie relative to the player,
%% as-matched?
%%
%% Returns: nomatch | {Xoffset, Yoffset}
%% Add Xoffset and Yoffset to the player's coordinates to find
%% the real position of element {X,Y} of the pattern.
%%
%% Pattern is a rectangular 2d list, each element is match character:
%%   E: our position - must occur exactly once
%%    : anything
%%   #: wall
%%   ~: water
%%   .: empty (can be adjacent to water)
%%   _: empty and not adjacent to water
%%   a: adjacent to water
%%   b: an enemy
%%   B: an enemy standing where we want to go next
%%   `: water, or adjacent to water
match(Pattern, MyId, MyNext, {X, Y}) ->
    %% We have to try 8 different patterns: turn the pattern sideways
    %% 1, 2, or 3 times, and then turn it upside down and try them all
    %% again.
    %% (We convert all coords to start from 0)
    X0 = X - 1,
    Y0 = Y - 1,
    W = length(hd(Pattern)),
    H = length(Pattern),
    case match_pattern(Pattern, X0, Y0, W, H, MyId, MyNext, 0) of
	nomatch ->
	    %% Try turning the pattern upside down
	    case match_pattern(reverse(Pattern),X0,H-Y0-1,W,H,MyId,MyNext,0) of
		nomatch ->
		    nomatch;
		Match ->
		    Match
	    end;
	Match ->
	    Match
    end.

%% Try to match with with 0-3 transpositions.
match_pattern(Pat0, X0, Y0, W0, H0, Id, Next, N) when N < 4 ->
    case match_1(Pat0, Id, Next) of
        true  ->
	    {Xoff, Yoff} = find_offsets(Pat0),
	    ?trace(io:format("Matched at ~p,~p (we are at ~p,~p)~nPattern:~n",
			     [X0, Y0, Xoff, Yoff])),
	    ?trace(foreach(fun(L) -> io:format("  ~p~n", [L]) end,
			   Pat0)),
	    %% Convert to our position relative to the player ({0,0}
	    %% is on him)
	    {X0 - Xoff, Y0 - Yoff};
        false ->
            Pat1 = transpose(Pat0),
            Y1 = W0 - X0 - 1,
            X1 = Y0,
	    W1 = H0,
	    H1 = W0,
            match_pattern(Pat1, X1, Y1, W1, H1, Id, Next, N+1)
    end;
match_pattern(_,_,_,_,_,_,_,_) ->
    nomatch.

%% Swap rows/columns in a 2D list (turn 90 degrees anti-clockwise).
transpose(L) -> transpose(L, []).

transpose([[]|_], Acc) ->
    Acc;
transpose(Rows, Acc) ->
    transpose([tl(R) || R <- Rows],
	      [[hd(R) || R <- Rows] | Acc]).

%% Match Pattern. Our robot id is Id, our next move is to Next.
match_1(Pattern, Id, Next) ->
    {{X,Y}, _} = bots_db:info_robot(Id),
    {Xoff, Yoff} = find_offsets(Pattern),
    Rec = #rec{id=Id, x0=X-Xoff, y0=Y-Yoff, pos={X,Y}, next=Next,
	       enemies=bots_db:all_other_robot_positions(Id)},
    patmatch(Pattern, Rec).

%% Find the {X,Y} position of the player ("E") relative to the
%% top-left corner of a pattern.
find_offsets([H|T]) -> find_offsets_1(H, T, 0, 0).

find_offsets_1([$E|T], Rows, X, Y)   -> {X, Y};
find_offsets_1([], [Row|Rows], X, Y) -> find_offsets_1(Row, Rows, 0, Y+1);
find_offsets_1([H|T], Rows, X, Y)    -> find_offsets_1(T, Rows, X+1, Y).

%% patmatch(Pattern, #rec) -> bool()
patmatch([Row|Rows], Rec = #rec{x0=X0, y0=Y0}) ->
    patmatch(Row, Rows, X0, Y0, Rec).
    
patmatch([], [], X, Y, Rec) ->
    true;
patmatch([], [P|Ps], X, Y, Rec) ->
    patmatch(P, Ps, Rec#rec.x0, Y+1, Rec);
patmatch([A|P], Ps, X, Y, Rec) ->
    case posmatch(A, X, Y, Rec) of
	true ->
	    patmatch(P, Ps, X+1, Y, Rec);
	false ->
	    false
    end.

posmatch(A, X, Y, #rec{pos={X,Y}}) ->
    %% at our position, only E matches
    A == $E;
posmatch($E, X, Y, Rec) ->
    false;
posmatch($b, X, Y, Rec) ->
    member({X,Y}, Rec#rec.enemies);
posmatch($B, X, Y, Rec) ->
    member({X,Y}, Rec#rec.enemies) and ({X,Y} == Rec#rec.next);
posmatch(A, X, Y, Rec) ->
    case {A, bots_db:info_cell({X, Y})} of
	{$#, wall}     -> true;
	{$~, water}    -> true;
	{$a, adjacent} -> true;
	{$., empty}    -> true;
	{$., adjacent} -> true;
	{$_, empty}    -> true;
	{$`, water}    -> true;
	{$`, adjacent} -> true;
	{$ , _}        -> true;
	_              -> false
    end.

corridor() ->
    ["#E#"].



