-module(bots_player).

-import(lists, [filter/2,map/2,foldl/3,member/2]).

% -define(debug,true).

-ifdef(debug).
-define(trace(E), E).
-else.
-define(trace(E), true).
-endif.

-record(ps, {id,				%Our id
	     cap,				%What we can carry
	     money,				%Cash
	     task=none,				%What we are doing
	     pos,				%Where WE think we should be
	     load=0,				%Load we are carrying
	     path=[],				%Path left to go
	     pkgs=[],				%Packages we are carrying
	     combat=none			%Combat state
						%none | {detour, OrigPath}
	    }).

%% Task = none |
%%        {collect,What} |
%%        {deliver,What}
%%
%% What = {Cost,Dest,Path}
%% Path = [Pos]
%% Pkgs = [{Cost,Dest,Path,[PkgId]}]

-export([init/3,reinit/1, turn/2]).

%% init(Id, Capacity, Money) -> State.

init(Id, Cap, M) ->
    {Cpos,Cpkgs} = bots_db:info_robot(Id),
    #ps{id=Id,cap=Cap,money=M,task=none,pos=Cpos,pkgs=[]}.

%% After a crash, reinit using the old state as a template
reinit(#ps{id=Id, cap=Cap, money=M}) ->
    init(Id, Cap, M).

%% turn(State, [LocPkgId]) -> {Bid,Move,State}.
%%  Compute one turn for the player. Here we put any special code for
%%  pushing or avoiding to be pushed, and also to handle the case of
%%  having been pushed.

turn(St, LocPkgs) ->
    {Bid0, St0} = combat(St),
    Bid = if Bid0 > St#ps.money ->
		  St#ps.money;
	     true ->
		  Bid0
	   end,
    %% Get current true data.
    {Cpos,Cpkgs} = bots_db:info_robot(St0#ps.id),
    %% If we are not where we think we are then we have been pushed.
    if Cpos /= St0#ps.pos ->
	    St1 = St0#ps{pos=Cpos},		%Store the new position
	    %% Cleanup after a push depends on our task.
	    case St1#ps.task of
		none ->				%Just update position
		    move(St1, LocPkgs, Bid);
		{collect,{OldC,Dest,OldPath}} ->
		    %% Keep same task but recompute.
		    {yes,{C,Path}} = bots_search:find_path(Cpos, Dest),
		    St2 = setup_path(Cpos, Path, St1),
		    move(St2, LocPkgs, Bid);
		{deliver,What} when Cpkgs /= [] ->
		    %% We still have some packages left to deliver
		    Ps = map(fun (Id) ->
				     {Pos,W,Dest,Stat} = bots_db:info_pkg(Id),
				     {Id,Dest,W}
			     end, Cpkgs),
		    {Cost,Dest,Path,Pkgs,W} = select_packages(Ps, St1),
		    bots_db:set_dest({Dest,W}),
		    St2 = setup_path(Cpos, Path, St1),
		    move(St2#ps{task={deliver,{Cost,Dest,Path}},pkgs=Pkgs},
			 LocPkgs,
			 Bid);
		{deliver,What} when Cpkgs == [] ->
		    %% Nothing left to deliver.
		    move(St1#ps{task=none,pkgs=[]}, LocPkgs, Bid)
	    end;
        true ->
	    %% We have not been pushed so keep on going.
	    move(St0, LocPkgs, Bid)
    end.

%% ------------------------------------------------------------
%% Combat

combat(St = #ps{id=Id, money=Money, combat={detour, OrigPath}}) ->
    %% We made a detour in the previous round. Now compute a new
    %% course, trying to avoid adjacent bad guys.
    if St#ps.task == none ->
	    %% No destination, nothing to compute
	    combat(St#ps{combat=none});
       true ->
	    {Pos, _} = bots_db:info_robot(Id),
	    Costs = adjacent_costs(Id),
	    Dest = lists:last(OrigPath),
	    ?trace(io:format("Recomputing course after detour: costs=~p dest=~p~n",
			     [Costs, Dest])),
	    {yes, {Cost,NewPath}} =
		bots_search:find_path_with_costs(Pos, Dest, Costs),
	    combat((setup_path(Pos, NewPath, St))#ps{combat=none})
    end;
combat(St0 = #ps{id=Id, path=Path, money=Money, task=Task}) ->

    %% DEBUG: slowdown to make the games easier to watch
    ?trace(receive after 500 -> ok end),

    Next = case Path of
	       [H|_] -> H;
	       _     -> undefined
	   end,
    %% Delivery is high-priority
    Bid = case {Path, Task} of
	      {[], {deliver, _}} ->
		  3;
	      _ ->
		  1
	  end,
    Bid0 = biggest(Bid, biggest(check_threat(Id, Next, Money),
				check_corridor(Id, Next, Money, St0#ps.task))),
    {Bid1,St1} =
	case check_kill(Id, Next, Money) of
	    {yes, {KillPos, KillBid}} ->
		{biggest(Bid0, KillBid),
		 St0#ps{path=[KillPos],combat={detour,Path}}};
	    no ->
		case check_evasion(Id, Next) of
		    {yes, EvPos} ->
			{biggest(Bid0, 3),
			 St0#ps{path=[EvPos],combat={detour,Path}}};
		    no ->
			{Bid0, St0}
		end
	end,
    {Bid1, St1}.

%% Give a list of {Pos,Cost} pricing for adjacent squares, based on
%% whether they're occupied.
adjacent_costs(Id) ->
    {{X,Y},_} = bots_db:info_robot(Id),
    Adj = [{X-1,Y}, {X+1,Y}, {Y-1,X}, {Y+1,X}],
    EnemyPositions = bots_db:all_other_robot_positions(Id),
    Occupied = [Pos || Pos <- Adj, member(Pos, EnemyPositions)],
    [{Pos, 3} || Pos <- Occupied].

%% Return the "biggest" of two bids. Judged by absolute value, except
%% that anything overrides 1 (i.e. -1 does).
biggest(1, B) ->
    B;
biggest(B1, B2) ->
    A1 = abs(B1),
    A2 = abs(B2),
    if A1 > A2 ->
	    B1;
       true ->
	    B2
    end.

%% Print a warning if there is a robot positioned to push us into water
check_threat(Id, Next, Money) ->
    case bots_match:match(["~Eb"], Id, Next, {1,1}) of
	{X,Y} ->
	    ?trace(io:format("DANGER: We can be pushed into the water (~p from here)~n",
			     [{X,Y}])),
	    lists:max([round(Money/10),
		       lists:min([2, Money])]);
	nomatch ->
	    1
    end.

%% Should we try and kill someone who's standing between us and some
%% water?
check_kill(Id, Next, Money) ->
    case bots_match:match(["~bE"], Id, Next, {2,1}) of
	{X,Y} ->
	    ?trace(io:format("MURDER: Let's kill the guy at ~p~n",
			     [{X,Y}])),
	    NRobots = length(bots_db:all_robots()),
	    {yes, {{X,Y},lists:max([round(Money/(10*NRobots)),
				    lists:min([2, Money])])}};
	nomatch ->
	    no
    end.

%% Print a warning if there is a robot where we want to go next, and a
%% non-watery place for us to side-step into.
check_evasion(Id, Next) ->
    case bots_match:match(["EB",
			   "_ "], Id, Next, {1,2}) of
	{X,Y} ->
	    ?trace(io:format("EVASION: Enemy in the way, we can side-step by ~p~n",
			     [{X,Y}])),
	    {yes, {X,Y}};
	nomatch ->
	    no
    end.

check_corridor(Id, Next, Money, Task) ->
    case bots_match:match(["#E#",
			   "#B#"], Id, Next, {2,2}) of
	{X,Y} ->
	    ?trace(io:format("CORRIDOR: We're facing an enemy in a confined space~n")),
	    case Task of
		{deliver, _} ->
		    %% We've got stuff: push on
		    3;
		_ ->
		    %% We've got nothing; back down
		    -2
	    end;
	nomatch ->
	    1
    end.

%% ------------------------------------------------------------
%% Strategy

%% move(State, [LocPkgId], Bid) -> {Move,State}.

move(#ps{task=none}=St0, [], Bid) ->	%Nothing to do
    bots_db:set_dest(undefined),
    %% Get all homes and convert into possible tasks.
    Ts = collect_positions(St0),
    case select_position(Ts, St0) of
	{C,Dest,Path} ->
	    St1 = setup_path(St0#ps.pos, Path, St0),
	    %%{Move,St2} = path_move(St1#ps.pos, St1),
	    %%{Bid,Move,pay_bid(St2#ps{task={collect,{C,Dest,Path}}}, Bid)};
	    move(St1#ps{task={collect,{C,Dest,Path}}}, [], Bid);
	none ->
	    %% No positions to go to, just idle.
	    {1,{pick,[-1]},pay_bid(St0, 1)}
    end;
move(#ps{task=none}=St0, LocPkgs, Bid) ->
    %% Use local packages for tasks.
    Ps = map(fun ({Id,X,Y,W}) -> {Id,{X,Y},W} end, LocPkgs),
    {Cost,Dest,Path,Pkgs,W} = select_packages(Ps, St0),
    bots_db:set_dest({Dest,W}),
    St1 = setup_path(St0#ps.pos, Path, St0),
    {Bid,{pick,Pkgs},
     pay_bid(St1#ps{task={deliver,{Cost,Dest,Path}},pkgs=Pkgs}, Bid)};
move(#ps{task={collect,What}}=St0, LocPkgs, Bid) ->
    %% Move towards a package pick up point.
    if  St0#ps.path == [] ->
	    %% We have now reached the pickup point, take what is there.
	    move(St0#ps{task=none}, LocPkgs, Bid);
	true ->
	    {Move,St1} = path_move(St0#ps.pos, St0),
	    {Bid,Move,pay_bid(St1, Bid)}
    end;
move(#ps{task={deliver,What}}=St0, LocPkgs, Bid) ->
    %% Move towards a package delivery point.
    if  St0#ps.path == [] ->
	    %% We have now reached the delivery point.
	    {Bid,{drop,St0#ps.pkgs},
	     pay_bid(St0#ps{task=none,pkgs=[]}, Bid)};
	true ->
	    {Move,St1} = path_move(St0#ps.pos, St0),
	    {Bid,Move,pay_bid(St1, Bid)}
    end.

%% collect_positions(State) -> {[Home],[PkgPos]}.
%%  Collect positions to go and collect packages.

collect_positions(St) ->
    {filter(fun (H) -> H /= St#ps.pos end, bots_db:all_homes()),
     map(fun (Id) ->
		 {Pos,W,Dest,Stat} = bots_db:info_pkg(Id),
		 Pos
	 end, bots_db:all_dropped())}.
    
%% select_position({[Home],[PkgPos]}, State) -> {Cost,Dest,Path} | none.
%%  Select a position from the homes and known packages.  We just pick
%%  the closest position.

select_position({Hs,Ps}, St) ->
    Cpos = St#ps.pos,
    Poss = ordsets:from_list(Hs ++ Ps),		%Get unique positions.
    Pos = foldl(fun (Pos, {Tc,Tpo,Tpa}=T) ->
			case bots_search:find_path(St#ps.pos, Pos) of
			    {yes,{Cost,Path}} when Cost < Tc ->
				{Cost,Pos,Path};
			    Else -> T		%No path, or higher cost
			end
		end, {none,Cpos,[Cpos,Cpos]}, Poss),	%None > int()
    case Pos of
	{none,_,_} -> none;
	Other -> Other
    end.

%% select_packages([{Id,Dest,W}], State) -> {Cost,Dest,Path,[PkgId],W}.
%%  Try to select the "best" packages.  We do this by picking the
%%  closest destination and tryin to take as many packages as possible
%%  to that destination.  We start with the lightest packages to try
%%  and take as many as possible.

select_packages(Ps, St) ->
    Db0 = orddict:new(),
    Db1  = foldl(fun ({Id,Pos,W}, Db) ->
			 orddict:append(Pos, {Id,W}, Db)
		 end, Db0, Ps),
    {Cost,Dest,Path,Pws0} =
	orddict:fold(fun (Dest, Pws, {Tc,Td,Tpa,Tpws}=T) ->
			     case bots_search:find_path(St#ps.pos, Dest) of
				 {yes,{Cost,Path}} when Cost < Tc ->
				     {Cost,Dest,Path,Pws};
				 Else -> T
			     end
		     end, {none,none,none,none}, Db1),
    Pws1 = lists:keysort(2, Pws0),		%Sort on weight
    {Pids,W} = select_packages_1(Pws1, St#ps.cap),
    {Cost,Dest,Path,Pids,W}.

select_packages_1([{Pid,W}|Pws], Cap) when W < Cap ->
    {Pids,Weight} = select_packages_1(Pws, Cap - W),
    {[Pid|Pids],Weight+W};
select_packages_1([Pw|Pws], Cap) -> {[],0};
select_packages_1([], Cap) -> {[],0}.

%%select_packages([{Id,Dest,W}|Ps], St) when W =< St#ps.cap ->
%%    {yes,{Cost,Path}} = bots_search:find_path(St#ps.pos, Dest),
%%    {Cost,Dest,Path,[Id],W}.

% select_packages(Ps0, St) ->
%     Db0 = orddict:new(),
%     Db1 = foldl(fun ({Id,Cost,Pos,Path}, Db) ->
% 			store(Pos, {Id,Cost,

%% setup_path(From, Path, State) -> State.
%%  Set up path move using the inital path from the from search (which
%%  contains initial position in the path).

setup_path(From, Path, St) ->
    St#ps{path=tl(Path)}.

%% path_move(From, State) -> {Move,State}.
%%  Make a move along the path updating our status.

path_move(Fr, St) ->
    Path = St#ps.path,
    To = hd(Path),
    Dir = move_dir(Fr, To),
    {{move,Dir},St#ps{pos=To,path=tl(Path)}}.

%% move_dir(From, To) -> Direction.
%%  Return direction to move.

move_dir({X1,Y1}, {X2,Y2}) when X1 < X2 -> east;
move_dir({X1,Y1}, {X2,Y2}) when X1 > X2 -> west;
move_dir({X1,Y1}, {X2,Y2}) when Y1 < Y2 -> north;
move_dir({X1,Y1}, {X2,Y2}) when Y1 > Y2 -> south.

pay_bid(St, Bid) ->
    St#ps{money=St#ps.money - Bid}.
