%%% File    : bots_search.erl
%%% Author  : Robert Virding
%%% Purpose : Path finding functions using A* algorithm.

-module(bots_search).

-import(lists, [reverse/1,member/2,map/2,flatmap/2,merge/2]).

-record(path, {e,c,p}).				%Estimated cost, cost, path

-export([find_path/2,find_path/3,find_path_with_costs/3,
	 astar_search/3,est_cost/2]).

%% find_path(Start, Dest) -> {yes,{Cost,Path}} | no.
%% find_path(Start, Dest, MaxCost) -> {yes,{Cost,Path}} | no.
%%  Find the best path to a place.  The cost returned is just the
%%  length of the path, users don't need to know HOW w arrived t our
%%  solution.

find_path(S, D) ->
    find_path(S, D, 1000000).

find_path(S, D, Max) ->
    find_path_1(S, D, Max, dict:new()).

%% Costs = [{Pos, Cost}] - additional cost information
find_path_with_costs(S, D, Costs) ->
    find_path_1(S, D, 1000000, dict:from_list(Costs)).

find_path_1(S, D, Max, Costs) ->
    case astar_search([#path{e=est_cost(S, D),c=0,p=[S]}], D, Max,
		     dict:new(), Costs) of
	[#path{e=E,c=C,p=P}|Ps] ->
	    {yes,{length(P),reverse(P)}};
	[] -> no
    end.
%% astar_search(Queue, Dest, MaxCost) -> {yes,RevPath} | no.
%%  This is the basic A* seach function.  It is relatively generically
%%  written but without going to extremes.  As an optimisation we keep
%%  a dictionary of visited positions and use this to determine if we
%%  should save a new path.

astar_search(Q, Dest, Max) ->
    astar_search(Q, Dest, Max, dict:new(), dict:new()).

astar_search([#path{}=P|Ps0], D, Max, Dict0, Costs) ->
    case hd(P#path.p) of
	D -> [P|Ps0];
	Other ->
	    Dict1 = dict:store(pos(P), cost(P), Dict0),
	    NewP = get_paths(P, D, Costs), %New valid non-looping paths
	    Ps1 = trim_merge(Ps0, NewP, Max, Dict1),
	    %%io:fwrite("~p~n", [{Ps0,Ps1}]),
	    astar_search(Ps1, D, Max, Dict1, Costs)
    end;
astar_search([], D, Max, Dict, Costs) -> []. %No path exists

%% trim_merge([Path], [NewPath], MaxCost, Dict) -> [Path].
%%  Merge in new paths into path list trimming away more costly paths
%%  to common nodes.  Use the dictionary of destinations to quickly
%%  determine if we should drop a new path.

%trim_merge(Ps, Ns, M, D) ->
%    lists:sort(fun comp_cost/2, Ns ++ Ps).

trim_merge(Ps, [N|Ns], Max, Dict) when N#path.c + N#path.e > Max ->
    trim_merge(Ps, Ns, Max, Dict);
trim_merge(Ps0, [N|Ns], Max, Dict) ->
    Cost = cost(N),
    case dict:find(pos(N), Dict) of
	{ok,OldCost} when OldCost =< Cost ->
	    %% The old one is better so forget the new one.
	    trim_merge(Ps0, Ns, Max, Dict);
	Else ->
	    %% First time or the new one is better.
	    Ps1 = trim_merge_1(Ps0, N),
	    trim_merge(Ps1, Ns, Max, Dict)
    end;
trim_merge(Ps, [], Max, Dict) -> Ps.

trim_merge_1([P|Ps], N) ->
    case comp_cost(N, P) of
	true -> [N,P|trim_merge_2(Ps, N)];
	false ->
	    case common_node(P, N) of
		true -> [P|Ps];
		false -> [P|trim_merge_1(Ps, N)]
	    end
    end;
trim_merge_1([], N) -> [N].

trim_merge_2([P|Ps], N) ->
    case common_node(P, N) of
	true -> Ps;
	false -> [P|trim_merge_2(Ps, N)]
    end;
trim_merge_2([], N) -> [].

comp_cost(#path{e=E1,c=C1,p=P1}, #path{e=E2,c=C2,p=P2}) ->
    C1 + E1 =< C2 + E2.

common_node(#path{p=[N1|_]}, #path{p=[N2|_]}) ->
    N1 == N2.

cost(#path{e=E,c=C}) -> E + C.

pos(#path{p=[H|_]}) -> H.

%% get_paths(Path, Dest) -> [Path].
%%  Get new paths from current paths. Remove loops and check for
%%  illegal postions.

get_paths(#path{c=C,p=P}, Dest, Costs) ->
    {X,Y} = hd(P),
    Adj = [{X-1,Y},{X,Y-1},{X,Y+1},{X+1,Y}],	%Very subtle!
    flatmap(fun (Squ) ->
		    %% Remove loops and illegal positions.
		    case member(Squ, P) of
			true -> [];		%A loop
			false ->
			    case legal_pos(Squ) of
				true ->
				    ExtraCost = case dict:find(Squ, Costs) of
						    error       -> 0;
						    {ok, Price} -> Price
						end,
				    [#path{e=est_cost(Squ, Dest),
					   c=C+pos_cost(Squ) + ExtraCost,
					   p=[Squ|P]}];
				false -> []
			    end
		    end
	    end, Adj).

est_cost({X1,Y1}, {X2,Y2}) ->
    abs(X1-X2) + abs(Y1-Y2).

pos_cost(Pos) ->
    case bots_db:info_cell(Pos) of
    %%case info_cell(Pos) of
	adjacent -> 5;				%Avoid water
	wall -> 1;
	water -> 1;
	empty -> 1
    end.

legal_pos(Pos) ->
    case bots_db:info_cell(Pos) of
    %%case info_cell(Pos) of
	empty -> true;
	adjacent -> true;
	water -> false;
	wall -> false
    end.

info_cell({2,3}) -> wall;
info_cell({2,4}) -> wall;
info_cell({2,5}) -> wall;
info_cell({2,7}) -> wall;
info_cell({2,8}) -> wall;
info_cell({2,9}) -> wall;
info_cell({2,10}) -> wall;
info_cell({2,11}) -> wall;
info_cell({2,12}) -> wall;
info_cell({3,3}) -> wall;
info_cell({4,2}) -> wall;
info_cell({4,3}) -> wall;
info_cell({5,2}) -> wall;
info_cell({6,2}) -> wall;
info_cell({6,1}) -> wall;
info_cell({3,6}) -> adjacent;
info_cell({4,6}) -> adjacent;
info_cell({5,3}) -> adjacent;
info_cell({5,4}) -> adjacent;
info_cell({5,5}) -> adjacent;
info_cell({5,6}) -> adjacent;
info_cell({3,4}) -> water;
info_cell({3,5}) -> water;
info_cell({4,4}) -> water;
info_cell({4,5}) -> water;
info_cell({X,Y}) when X < 1 -> wall;
info_cell({X,Y}) when Y < 1 -> wall;
info_cell({X,Y}) -> empty.
